
/**************************************************************************
 *
 *  $Id: log_adj.c 1.8 2012/05/23 13:35:41Z martin REL_M $
 *  $Name: LOG_ADJ_1_6 $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Detect system time corrections on Windows platforms.
 *
 * -----------------------------------------------------------------------
 *  $Log: log_adj.c $
 *  Revision 1.8  2012/05/23 13:35:41Z  martin
 *  Support precision time API introduced with Windows 8.
 *  Revision 1.7  2012/05/15 10:35:08Z  martin
 *  Support building (also x64 version) under VS2008 Express
 *  which does not accept the VC6-generated resource records.
 *  Revision 1.6  2009/02/05 09:21:39Z  martin
 *  Picked up new command line options implemented by Dave Hart:
 *  -q  Sample QueryPerformanceCounter before time
 *  -s   Use Sleep() instead of CreateWaitableTimer()
 *  Use waitable timer now by default.
 *  Revision 1.5  2006/04/28 14:28:52Z  martin
 *  Support command line parameters, print usage.
 *  Write to file only optionally.
 *  Optionally modify MM timer while active.
 *  Revision 1.4  2006/04/28 07:46:04Z  martin
 *  Compute deltas as double.
 *  Show current perf count conditionally only.
 *  Revision 1.3  2004/04/15 13:37:05Z  martin
 *  Print performance counter value.
 *  Print version info retrieved from the version resource which 
 *  has been addedto the executable file.
 *  Revision 1.2  2003/10/07 15:24:22Z  martin
 *  Check and display deltas of system time and performance
 *  counter in order to detect time steps.
 *  Added standard file header.
 *
 **************************************************************************/

#ifndef _WIN32_WINNT
  #define _WIN32_WINNT 0x0400
#endif

#define _CRT_SECURE_NO_WARNINGS

#include <prectime.h>

#include <windows.h>
#include <stdio.h>
#include <time.h>

#define SHOW_CURR_PERF_COUNT 0

#define LIM             20
#define MM_TIMER_INTV   1  /* the interval we'd want to set the MM timer to [ms] */


static const char ver_str[] = "1.6";


typedef union
{
  FILETIME ft;
  LONGLONG ll;
} LARGE_FILETIME;


typedef struct
{
  LARGE_FILETIME lft;
  LARGE_INTEGER perf_cnt;
} LARGE_FILETIME_PERF_CNT;


static BOOLEAN must_print_usage;
static BOOLEAN modify_mm_timer;
static BOOLEAN qpf_first;
static BOOLEAN use_sleep;
static BOOLEAN use_gstapft;

static UINT wTimerRes;
static TIMECAPS tc;
static char *fn;

static GSTAFT_FNC gstaft_fnc;
static GSTAFT_FNC gstapft_fnc;



/*HDR*/
char *sprint_win_file_version( char *s )
{
  char tmp_fn[MAX_PATH];
  int success = 0;
  LPVOID p_vs_info;
  VS_FIXEDFILEINFO *p;
  DWORD dw;
  DWORD sz;
  UINT len;


  if ( GetModuleFileName( NULL, tmp_fn, sizeof( tmp_fn ) ) == 0 )
    goto done;


  sz = GetFileVersionInfoSize( tmp_fn, &dw );

  if ( sz == 0 )
  {
    // we couldn't access the resource record, so we
    // can't get the fersion number from the resources
    // and thus use the default string
    strcpy( s, ver_str );
    success = 1;
    goto done;
  }


  p_vs_info = malloc( sz );

  if ( p_vs_info )
  {
    if ( GetFileVersionInfo( tmp_fn, (DWORD) 0, sz, p_vs_info ) )
    {
      if ( VerQueryValue(
             p_vs_info,     // buffer for version resource
             TEXT( "\\" ),  // value to retrieve
             &p,            // buffer for version value pointer
             &len           // version length
           ) )
      {
        sprintf( s, "%u.%u", 
                 p->dwProductVersionMS >> 16,
                 p->dwProductVersionMS & 0xFFFF );

        if ( p->dwProductVersionLS )
        {
          sprintf( &s[strlen( s ) ], ".%u.%u", 
                   p->dwProductVersionLS >> 16,
                   p->dwProductVersionLS & 0xFFFF );
        }

        if ( strcmp( s, ver_str ) )
          strcat( s, " (!)" );
 
        success = 1;
      }
    }

    free( p_vs_info );
  }


done:
  if ( !success )
    s[0] = 0;

  return s;

}  // sprint_win_file_version



static /*HDR*/
void log_to_file( const char *s )
{
  FILE *fp;

  if ( fn )
  {
    fp = fopen( fn, "at" );

    if ( fp )
    {
      fprintf( fp, "%s\n", s );
      fclose( fp );
    }
  }

}  // log_to_file



static /*HDR*/
void log_msg( const char *s )
{
  printf( "%s\n", s );
  log_to_file( s );

}  // log_msg



static /*HDR*/ /* atexit fnc */
void restore_mm_timer( void )
{
  if ( wTimerRes )  /* if not 0 then the MM timer has been modified at startup */
  {
    timeEndPeriod( wTimerRes ); 
    wTimerRes = 0;
  }

}  // restore_mm_timer



static /*HDR*/ /* atexit fnc */
void free_fn( void )
{
  if ( fn )
  {
    free( fn );
    fn = NULL;
  }

}  // free_fn



BOOL WINAPI
OnConsoleEvent( DWORD dwCtrlType )
{
  switch ( dwCtrlType )
  {
    case CTRL_BREAK_EVENT:
    case CTRL_C_EVENT:
    case CTRL_CLOSE_EVENT:
    case CTRL_SHUTDOWN_EVENT:
      exit( 0 );

    default:
      return FALSE;
  }

  return TRUE;

}  // OnConsoleEvent



static
void get_times( LARGE_FILETIME_PERF_CNT *p )
{
  if ( !qpf_first )
  {
    gstaft_fnc( &p->lft.ft );
    QueryPerformanceCounter( &p->perf_cnt );
  }
  else
  {
    QueryPerformanceCounter( &p->perf_cnt );
    gstaft_fnc( &p->lft.ft );
  }

}  // get_times



int main( int argc, char *argv[] )
{
  char ver_str[80];
  char ws[256];
  LARGE_FILETIME_PERF_CNT prv;
  LARGE_FILETIME_PERF_CNT curr;
  double DeltaFileTime;
  SYSTEMTIME CalTime;
  LARGE_INTEGER PerfFreq;
  LARGE_INTEGER DueTime;
  double DeltaPerfCnt;
  DWORD StdTimeAdjustment;
  DWORD CurrTimeAdjustment;
  double DeltaT;
  BOOL disabled;
  int i;
  HANDLE Timer = INVALID_HANDLE_VALUE;
  DWORD rc;

  sprintf( ver_str, "Meinberg LOG_ADJ v%s", sprint_win_file_version( ws ) );
  printf( "\n%s\n", ver_str );


  if ( !QueryPerformanceFrequency( &PerfFreq ) )
  {
    log_msg( "Unable to read perfomance counter frequency." );
    exit( 1 );
  }

	if ( !SetConsoleCtrlHandler( OnConsoleEvent, TRUE ) )
	{
		log_msg( "Unable to set console control handler.");
    exit( 1 );
	}

  // check program arguments
  for ( i = 1; i < argc; i++ )
  {
    char *cp = argv[i];

    if ( ( strcmp( cp, "-?" ) == 0 )       // help
      || ( strcmp( cp, "-h" ) == 0 ) )
    {
      must_print_usage = TRUE;
    }
    else if ( strcmp( cp, "-o" ) == 0 )    // output to specified file
    {
      int next_arg_idx = i + 1;

      if ( next_arg_idx < argc )
      {
        fn = argv[next_arg_idx];
        i++;
      }
      else
        must_print_usage = TRUE;
    }
    else if ( strcmp( cp, "-l" ) == 0 )    // output to newly generated file
    {
      time_t t = time( NULL );
      sprintf( ws, "%08lx.txt", (unsigned long) t );
      fn = malloc( strlen( ws ) + 1 );
      if ( fn )
      {
        atexit( free_fn );
        strcpy( fn, ws );
      }
      else
      {
      }
    }
    else if ( strcmp( cp, "-m" ) == 0 )    // modify MM timer
    {
      modify_mm_timer = TRUE;
    }
    else if ( strcmp( cp, "-p" ) == 0 )    // use precise FILETIME, if supported by the OS
    {
      use_gstapft = TRUE;
    }
    else if ( strcmp( cp, "-q" ) == 0 )    // sample QPF before system time
    {
      qpf_first = TRUE;
    }
    else if ( strcmp( cp, "-s" ) == 0 )    // use Sleep()
    {
      use_sleep = TRUE;
    }
    else                                   // unknown parameter
    {
      sprintf( ws, "unknown parameter: %s\n", cp );
      must_print_usage = 1;
    }
  }

  if ( must_print_usage )
  {
    printf( "\n"
            "usage:\n"
            "\n"
            "log_adj [options]\n"
            "\n"
            "with the optional parameters:\n"
            "\n"
            "  -l        Log output to a file nnnnnnnn.txt, where\n"
            "            nnnnnnnn is a hexadecimal representation of\n"
            "            the current time in seconds since 1970\n"
            "\n"
            "  -o fname  Log output to file named fname\n"
            "\n"
            "  -m        Set Windows multimedia timers to highest resolution\n"
            "            while running\n"
            "  -q        Sample QueryPerformanceCounter before time\n"
	          "  -s        Use Sleep() instead of CreateWaitableTimer()\n"
            "  -p        Use GetSystemTimePreciseAsFileTime() if supported\n"
            "\n" );
    exit( 1 );
  }


  // Here program execution continues normally. 
  // First write our version information to file, if required.
  log_to_file( ver_str );


  // By default use GetSystemTimeAsFileTime() to read the current time
  gstaft_fnc = GetSystemTimeAsFileTime;

  // Check if the new API GetSystemTimePreciseAsFileTime() is supported,
  // and if we want to use it.
  rc = import_precise_time_api( &gstapft_fnc );

  if ( use_gstapft )
  {
    if ( gstapft_fnc )
    {
      gstaft_fnc = gstapft_fnc;
      sprintf( ws, "Using GetSystemTimePreciseAsFileTime API\n" );
      log_msg( ws );
    }
    else
    {
      sprintf( ws, "GetSystemTimePreciseAsFileTime API not supported, falling back to default\n" );
      log_msg( ws );
    } 
  }
  else
  {
    if ( gstapft_fnc )
    {
      sprintf( ws, "Note: GetSystemTimePreciseAsFileTime API is also available, use -p\n" );
      log_msg( ws );
    }
  }

  if ( modify_mm_timer )
  {
    if ( timeGetDevCaps( &tc, sizeof( tc ) ) == TIMERR_NOERROR ) 
    {
      wTimerRes = min( max( tc.wPeriodMin, MM_TIMER_INTV ), tc.wPeriodMax );

      timeBeginPeriod( wTimerRes );
      atexit( restore_mm_timer );
      sprintf( ws, "MM timer resolution: %u..%u ms, set to %u ms",
               tc.wPeriodMin, tc.wPeriodMax, wTimerRes );
    }
    else
      sprintf( ws, "Failed to get MM timer caps." );

    log_msg( ws );
  }

  log_msg( "" );

  get_times( &prv );

  if ( !use_sleep )
  {
    Timer = CreateWaitableTimer( NULL, FALSE, NULL );
    DueTime.QuadPart = 0;
    SetWaitableTimer( Timer, &DueTime, 1000, NULL, NULL, FALSE );
    // first wait returns instantly, curiously
    WaitForSingleObject( Timer, INFINITE );
  }

  for (;;)
  {
    if ( use_sleep )
    {
      Sleep( 1000 );
    }
    else
    {
      WaitForSingleObject( Timer, INFINITE );
    }

   
    get_times( &curr );

    GetSystemTimeAdjustment( &CurrTimeAdjustment, &StdTimeAdjustment, &disabled );

    FileTimeToSystemTime( &curr.lft.ft, &CalTime );

    DeltaFileTime = (double) ( curr.lft.ll - prv.lft.ll ) / 10000.0;
    DeltaPerfCnt = ( (double) ( curr.perf_cnt.QuadPart - prv.perf_cnt.QuadPart ) * 1000.0 )
                   / (double) PerfFreq.QuadPart;
    DeltaT = DeltaFileTime - DeltaPerfCnt;
    
    sprintf( ws, "%04u-%02u-%02u %02u:%02u:%02u.%04u: "
             "%9.4f %9.4f %+9.4f  "
             "%i (%+i) "
             #if SHOW_CURR_PERF_COUNT
               "%I64u "
             #endif
             "%s", 
             CalTime.wYear, CalTime.wMonth, CalTime.wDay,
             CalTime.wHour, CalTime.wMinute, CalTime.wSecond, CalTime.wMilliseconds,

             DeltaFileTime,
             DeltaPerfCnt,
             DeltaT,

             CurrTimeAdjustment, 
             CurrTimeAdjustment - StdTimeAdjustment,
             #if SHOW_CURR_PERF_COUNT
               (__int64) CurrPerfCnt.QuadPart,
             #endif
             ( DeltaT > LIM || DeltaT < -LIM ) ? "><" : ""
           );

    log_msg( ws );

    prv = curr;
  }
}